define('confluence-collaborative-editor-plugin/synchrony-content', [
    'jquery',
    'underscore',
    'ajs',
    'confluence/meta',
    'confluence/legacy',
    'confluence-collaborative-editor-plugin/synchrony-util'
], function (
    $,
    _,
    AJS,
    Meta,
    Confluence,
    Util
) {
    'use strict';

    var lastKnownTitle;

    function getConfRev(content) {
        return content.confRev || Util.retrieveMetadata('conf-revision');
    }

    function getSyncRev(content) {
        var syncRev = content.syncRev || $('#syncRev').val() || null;
        return syncRev === 'dummy-sync-rev' ? null : syncRev;
    }

    function getPageTitle() {
        var pageTitle = Meta.get('page-title') || Meta.get('original-content-title');
        if (null === pageTitle || typeof pageTitle === 'undefined') {
            pageTitle = '';
        }
        return pageTitle;
    }

    function toSynchronyFormat(title, editorContent) {
        return "<body data-title='" + _.escape(title) + "'>" + editorContent + "</body>";
    }

    function hasError(editorContent) {
        // TODO Is there a better way?
        return !!$('<div>' + editorContent + '</div>').find('.fatal-render-error').length;
    }

    // TODO Remove innerHTML once Confluence triggers the 'rte-content-ready' event
    function getEditorContent(content) {
        if (content.editorContent) {
            return content.editorContent;
        }

        var editorContent = $('.rte-content-ready');
        if (!!editorContent.length) {
            return editorContent.val();
        }

        return AJS.Rte.getEditor().getBody().innerHTML; // innerHTML kept for backwards compatibility
    }

    function getContent(data) {
        var content = data || {};
        var title = content.title || getPageTitle();
        var editorContent = getEditorContent(content);

        return {
            title: title,
            raw: editorContent,
            html: toSynchronyFormat(title, editorContent),
            error: hasError(editorContent),
            syncRev: getSyncRev(content),
            confRev: getConfRev(content)
        };
    }

    function isUnpublished() {
        return Util.retrieveMetadata('new-page');
    }

    function isSynchronyContainerTheFirstElement(htmlDoc) {
        // if the first element is the synchrony container it means that there are no valid tinymce caret containers
        return htmlDoc.firstChild && htmlDoc.firstChild.classList && htmlDoc.firstChild.classList.contains('synchrony-container');
    }

    function doesNotHaveAValidCaretContainer(htmlDoc) {
        return !htmlDoc.children.length || (htmlDoc.children.length === 1 && isSynchronyContainerTheFirstElement(htmlDoc));
    }

    function fixTinymceCaretContainer(htmlDoc, editor) {
        if (doesNotHaveAValidCaretContainer(htmlDoc)) {
            // tinymce setContent when supplied with an empty string will default to
            // include an empty paragraph with a bogus br for some browsers or a space for others
            // this will ensure we always have a tinymce caret container
            editor.setContent('');
        }
    }

    function writeTitleToRootElement() {
        var title = $('#content-title').val();
        if (title !== lastKnownTitle) {
            var docFrame = frames['wysiwygTextarea_ifr'];
            var document = docFrame.contentDocument !== undefined ? docFrame.contentDocument : docFrame.document;
            document.body.setAttribute('data-title', title);
            lastKnownTitle = title;
        }
    }

    function readTitleFromRootElement(body) {
        if (body.hasAttribute('data-title')) {
            var title = body.getAttribute('data-title');
            if (title !== lastKnownTitle) {
                $('#content-title').val(title);
                lastKnownTitle = title;
            }
        }
    }

    function bindPostPasteFix() {
        var $doc = $(document);
        var synchronyId = 'confluence.postpaste-fix';

        $doc.bind('prePaste', function() {
            AJS.trigger('synchrony.stop', {
                id: synchronyId
            });
        }).bind('postPaste', function() {
            setTimeout(function() {
                AJS.trigger('synchrony.start', {
                    id: synchronyId
                });
            }, 0);
        });
    }

    function isContentEmpty(rawContent) {
        // empty if just whitespace or html with empty text nodes
        return /^\s+$/.test(rawContent) || $(rawContent).text().trim().length === 0;
    }

    return {
        getContent: getContent,
        isUnpublished: isUnpublished,
        fixTinymceCaretContainer: fixTinymceCaretContainer,
        writeTitleToRootElement: writeTitleToRootElement,
        readTitleFromRootElement: readTitleFromRootElement,
        bindPostPasteFix: bindPostPasteFix,
        isContentEmpty: isContentEmpty
    };
});
