define('confluence-collaborative-editor-plugin/synchrony-editor', [
    'jquery',
    'ajs',
    'confluence/legacy',
    'confluence-collaborative-editor-plugin/editor-blanket',
    'confluence-collaborative-editor-plugin/synchrony-util',
    'confluence-collaborative-editor-plugin/synchrony-content',
    'confluence-collaborative-editor-plugin/synchrony-auth',
    'confluence-collaborative-editor-plugin/synchrony-entity',
    'confluence-collaborative-editor-plugin/synchrony-handlers'
], function(
    $,
    AJS,
    Confluence,
    EditorBlanket,
    Util,
    Content,
    Auth,
    SynchronyEntity,
    Handlers) {

    'use strict';

    var exports = {};

    Util.time("confluence.editor.synchrony.jsLoad");

    var SYNCHRONY_CONNECTION_TIMEOUT_MS = 30 * 1000; // 30s

    var statusIndicator;

    var didSynchronyConnect = false;
    var didUserHitTimeout = false;

    $(window).bind("beforeunload", function() {
        if (!didSynchronyConnect) {
            AJS.trigger('analyticsEvent', {
                name: 'confluence.synchrony.exit.before.connecting',
                data: {
                    afterTimeout: didUserHitTimeout,
                    contentId: Confluence.getContentId()
                }
            });
        }
    });

    // For page objects to wait correctly in tests
    function setCollaborativeEditorStatus(status) {
        $('meta[name="ajs-collaborative-editor-status"]').attr('content', status);
    }

    function isEditorInitialised() {
        return AJS.Rte.getEditor() && AJS.Rte.getEditor().initialized;
    }

    function isSlowEdit() {
        return !AJS.Rte.isQuickEdit
    }

    // TODO Can this be moved to the status indicator file?
    function loadStatusIndicator() {
        var statusIndicatorReady = $.Deferred();

        AJS.$(function() {
            AJS.$('body').addClass('synchrony-active');
            var StatusIndicator = require('confluence-collaborative-editor-plugin/status-indicator-view');
            statusIndicator = new StatusIndicator({
                el: '#pluggable-status',
                model: new Backbone.Model({
                    minActiveTime: 500
                })
            });

            Auth.init(statusIndicator);
            statusIndicatorReady.resolve();
        });
        return statusIndicatorReady.promise();
    }

    function showLoadFailureMessage() {
        AJS.trigger('editor.error.message', {
            messageKey: 'collaborative-editor-load-failure',
            message: AJS.I18n.getText('editor.collaborative-editing.load.error'),
            disablePublish: true,
            close: 'manual'
        });
    }

    function hideLoadFailureMessage() {
        AJS.trigger('dismiss.editor.error.message', {
            messageKey: 'collaborative-editor-load-failure',
            enablePublish: true
        });
    }

    function alwaysResolve(promise, label) {
        var d = $.Deferred();
        promise.done(function(){Util.timeEnd(label);}, d.resolve);
        promise.fail(d.resolve);
        return d.promise();
    }

    exports._bindEditor = function (content, editor, snapshotRes, contentReconciliationRes) {
        Util.time("confluence.editor.synchrony.init");
        // The condition where we trigger a resynchronisation of confluence and synchrony will be an invalid ancestor
        // error from the content reconciliation request which will occur whether they know about the entity or not.
        var crResponseContent = ((contentReconciliationRes[1] === 'error' && contentReconciliationRes[0].responseText)
            ? JSON.parse(contentReconciliationRes[0].responseText)
            : "");

        var behind = false;
        if (crResponseContent.type === 'invalid-ancestor') {
            behind = "synchrony";
        } else if (crResponseContent.type === 'out-of-order-revision') {
            behind = "confluence";
        }

        if(behind) {
            AJS.log('Performing synchrony recovery');
            $.ajax({
                url: AJS.contextPath() + '/rest/synchrony/1.0/content/' + Confluence.getContentId() + '/recovery?behind=' + behind + '&conflictingRev=' + crResponseContent['conflicting-rev'],
                type: 'PUT'
            });
            return null; // error
        }

        if ((contentReconciliationRes[1] !== "success" && snapshotRes[1] !== "success") ||
            (contentReconciliationRes[1] !== "success" && crResponseContent.type !== "duplicate-mismatch")) {
            // client-side content-reconciliation has failed
            AJS.trigger("analyticsEvent", {
                name: "confluence.synchrony.client.content.reconciliation.failure",
                data: {
                    type: crResponseContent.type,
                    contentId: Confluence.getContentId()
                }
            });
            return null; // error
        }

        var $revField = $('#syncRev');

        // If we have a snapshot then we want to use it as the base state of the
        // entity. 'history.base' is the state the entity will materialize into
        // the dom upon initialization.
        var history = null;
        if (snapshotRes[1] === 'success') {
            var snapshotInfo = snapshotRes[0];
            history = {base: {rev: Synchrony.rev(snapshotInfo.stateAt),
                state: snapshotInfo.state}};
            var deltaInfo = contentReconciliationRes[0];
            if (contentReconciliationRes[1] === 'success'
                && snapshotInfo.stateAt === deltaInfo.deltaFrom) {
                history.base.rev = Synchrony.rev(deltaInfo.deltaTo);
                history.base.state = Synchrony.materialize(
                    history.base.state,
                    Synchrony.unmarshal(deltaInfo.delta.value)
                );
            }
            $revField.val(history.base.rev.toString());
        }

        AJS.log('Synchrony will load with contentId: ', Confluence.getContentId());

        try {
            // Ensure the document the user starts out with includes the
            // same title attribute as the one in the CR request,
            // otherwise the incoming changes from the CR request in the
            // connected event will overwrite any changes to the title
            // after the blanket is lifted and before the connected
            // event.
            var htmlDoc = editor.getBody();
            htmlDoc.setAttribute('data-title', content.title);

            var entity = SynchronyEntity.bind(Synchrony, editor, htmlDoc, history);

            Handlers.handle(entity, editor, statusIndicator);

            /**
             * when synchrony enity initialize, any content which was set to editor is cleared,
             * and synchrony append snapshot as the content for editor
             * we trigger remote change so any tinymce plugin can run through init process again on this new content
             * then set editor to not dirty
             */
            AJS.trigger('editor.remote.change');
            editor.setDirty(false);
            Util.timeEnd("confluence.editor.synchrony.init");

            return entity;
        } catch (e) {
            AJS.logError(e);
            return null; // error
        }
    };

    // Fetches the head state of the document from synchrony.
    function requestSnapshot(currentToken) {
        function snapshotRequestProvider(token) {
            Util.time("confluence.editor.synchrony.snapshot");

            return alwaysResolve($.ajax({
                type: 'GET',
                headers: {'x-token': token},
                url: encodeURI(Util.getServiceUrl() + '/data' + Util.getEntityId() + '?state-at=@head&state-format=type-tagged'),
                dataType: 'json'
            }).pipe(Util.asArray, Util.asArray), "confluence.editor.synchrony.snapshot");
        }

        return Auth.performRequest(snapshotRequestProvider, currentToken);
    }

    // Reconciles the published page with synchrony or creates the document from the given content if it doesn't yet exist.
    function requestContentReconciliation(content, currentToken, deltaFrom) {
        function crRequestProvider(token) {
            Util.time("confluence.editor.synchrony.CR");

            var url = Util.getServiceUrl() + '/data' + Util.getEntityId() + '?optimistic=true';
            if (deltaFrom) {
                url += '&delta-from=' + deltaFrom + '&delta-to=@head&delta-format=type-tagged';
            }

            return alwaysResolve($.ajax({
                type: 'PUT',
                headers: {'x-token': token},
                url: encodeURI(url),
                dataType: 'json',
                contentType: 'application/json',
                data: JSON.stringify({
                    ancestor: content.syncRev,
                    rev: content.confRev,
                    state: {
                        format: 'html',
                        value: content.html
                    },
                    merges: {master: {meta: {type: 'client-reconciliation'}}}
                })
            }).pipe(Util.asArray, Util.asArray), "confluence.editor.synchrony.CR");
        }

        // Don't do CR when content completely empty for new pages CONFDEV-47057
        if (!content.syncRev && Content.isUnpublished() && (content.raw.length === 0 || /^\s+$/.test(content.raw))) {
            return $.Deferred().resolve([{}, "success"]);
        } else if (Content.isContentEmpty(content.raw)) {
            AJS.trigger('analyticsEvent', {
                name: 'confluence.synchrony.client.content.reconciliation.blank-content',
                data: {
                    contentId: Confluence.getContentId()
                }
            });
            return $.Deferred().resolve([{}, "success"]);
        }
        return Auth.performRequest(crRequestProvider, currentToken);
    }

    function trackEditorReady(collabEditorReady) {
        Util.time("confluence.editor.synchrony");

        $.when(Date.now(), collabEditorReady).pipe(function (startTime) {
            setCollaborativeEditorStatus('on');

            var endTime = Date.now();
            AJS.trigger('analyticsEvent', {
                name: 'confluence.synchrony.editor.loaded',
                data: {
                    durationMillis: (endTime - startTime),
                    contentId: Confluence.getContentId()
                }
            });

            AJS.trigger("rte-collab-ready");

            Util.timeEnd("confluence.editor.synchrony");
            Util.timeEnd("confluence.editor");
        });
    }

    function trackConnection() {
        // synchrony.connected is redundant due to how namespaced events work, but let's be explicit
        // or change the synchrony.connected.fake at some point (triggered from webdriver page objects)
        var synchronyConnected = makePromise('synchrony.connected synchrony.connected.fake');
        var firstConnection = makePromise('synchrony.connected synchrony.connected.fake synchrony-unsupported-extension', SYNCHRONY_CONNECTION_TIMEOUT_MS);

        // A timeout here could be for a number of reasons. Plugin not configured correctly, plugin disabled,
        // or a failure to connect to synchrony, or a very long response from synchrony.
        firstConnection.fail(function () {
            showLoadFailureMessage();
            AJS.trigger('analyticsEvent', {
                name: 'confluence.synchrony.editor.load.timeout',
                data: {
                    contentId: Confluence.getContentId()
                }
            });

            didUserHitTimeout = true;

            // Connection recovered after a while
            synchronyConnected.done(function() {
                didSynchronyConnect = true;
                hideLoadFailureMessage();
            });
        });

        synchronyConnected.done(function() {
            didSynchronyConnect = true;
        });
    }

    exports._initialiseCollabEditingComponents = function() {
        var indicatorPromise = loadStatusIndicator();
        // Needed for webdriver tests that run without a running synchrony service,
        // which would otherwise never have the blanket lifted.
        $(window).one('synchrony.connected.fake', function (e) {
            if (e.namespace === 'connected.fake') { //Namespaced events FTW
                hideLoadFailureMessage();
                setCollaborativeEditorStatus('fake');
            }
        });
        return indicatorPromise;
    };

    exports._initialiseSynchrony = function () {
        Util.time("confluence.editor.synchrony.deps");
        // Synchrony's async loader API, copy&pasted from the Synchrony API documentation.
        (function(b){var a=b.Synchrony=b.Synchrony||{};if(!a.ready){var e=a._cbs=a._cbs||[];a.ready=function(a){e.push(a)}}var c=function(){1===a.state&&(a.SockJS=b.SockJS,a.init())};if(b.SockJS)a.init&&c();else{var d=b._sockjs_onload;b._sockjs_onload=function(){c();d&&d.apply(this,arguments)}}})(window);

        var synchronyReadyDeferred = $.Deferred();
        Synchrony.ready(synchronyReadyDeferred.resolve);

        return $.when(
            synchronyReadyDeferred.promise(),
            // wait for script loading errors
            Util.loadScript('/js/vendor/sockjs.min.js'),
            Util.loadScript('/js/synchrony.min.js')
        ).pipe(function (Synchrony) {
            Util.timeEnd("confluence.editor.synchrony.deps");
            AJS.log('Synchrony successfully initialised.');
            return Synchrony;
        });
    };

    function initialiseCollabEditing(contentLoaded, collabEditorLoaded) {
        var synchronyReady = exports._initialiseSynchrony();
        var collabEditorReady = makePromise('synchrony.connected.fake');

        exports._initialiseCollabEditingComponents().pipe(function() {
            collabEditorLoaded.pipe(function () {
                trackConnection();
                trackEditorReady(collabEditorReady);
                EditorBlanket.applyBlanket(collabEditorReady);
            });
            contentLoaded.pipe(function (content) {
                // The content may be loaded a long time after the page has loaded, so make sure we have a valid not expired token
                return $.when(content, Auth.getTokenPromise());
            }).pipe(function (content, token) {
                var snapshotReq = $.when(requestSnapshot(token), synchronyReady).pipe(function (snapshotRes, Synchrony) {
                    if (snapshotRes[1] === 'success') {
                        return [{
                            stateAt: snapshotRes[0].stateAt,
                            state: Synchrony.unmarshal(snapshotRes[0].state.value)
                        }, snapshotRes[1]];
                    } else {
                        return snapshotRes;
                    }
                });
                var contentReconciliationReq = requestContentReconciliation(content, token, null);
                return $.when(content, collabEditorLoaded, snapshotReq, contentReconciliationReq);
            }).pipe(function (content, editor, snapshotRes, contentReconciliationRes) {
                var entity = exports._bindEditor(content, editor, snapshotRes, contentReconciliationRes);
                if (!entity) {
                    AJS.trigger("analyticsEvent", {
                        name: "confluence.synchrony.bind.failure",
                        data: {
                            contentId: Confluence.getContentId()
                        }
                    });
                    return $.Deferred().reject('editor-binding-failed');
                }
                collabEditorReady.resolve();
            }).fail(function (error) {
                collabEditorLoaded.always(function () { //We need to wait until the editor is loaded to actually show the errors, etc.
                    if (error !== 'synchrony-not-enabled') {
                        AJS.log('Failed to load the collaborative editor', error);
                        showLoadFailureMessage();
                        AJS.$(function () {
                            statusIndicator.onDisconnectedState();
                        });
                        setCollaborativeEditorStatus('failed');
                        AJS.trigger("analyticsEvent", {
                            name: "confluence.synchrony.error",
                            data: {
                                error: error,
                                contentId: Confluence.getContentId()
                            }
                        });
                    }
                });
            });
        });
    }

    exports.init = function () {
        // TODO Remove rte.init.ui once Confluence triggers the 'rte-content-ready' event, as it is triggered earlier
        var collabEditingEnabled = makeInitPromise('rte.init.ui rte-content-ready'); //rte.init.ui kept for backwards compatibility

        var contentLoaded = makeInitPromise('rte-content-ready rte-quick-edit-ready', Content.getContent);
        contentLoaded = makeInitPromise('rte-ready', Content.getContent, isSlowEdit, contentLoaded);

        var collabEditorLoaded = makeInitPromise('rte-quick-edit-ready', AJS.Rte.getEditor);
        collabEditorLoaded = makeInitPromise('rte-ready', AJS.Rte.getEditor, isSlowEdit, collabEditorLoaded);

        collabEditingEnabled.pipe(function () {
            initialiseCollabEditing(contentLoaded.promise(), collabEditorLoaded.promise());
        });

        collabEditorLoaded.fail(function (error) {
            if (error === 'synchrony-not-enabled') {
                nonCollabEditingShenanigans();
                AJS.bind('rte-ready', nonCollabEditingShenanigans); // Needed for comments (might be triggered multiple times)
                AJS.log('Synchrony is not available in this context.');
            }
        });
    };

    // The editor is by default hidden. Collaborative draft-edit mode is handled by the Synchrony init process.
    // For any other editors/modes (comments), or when Synchrony is not enabled or can't be enabled,
    // we still have to show the editor every time it becomes ready (which may be multiple times on a single page).
    function nonCollabEditingShenanigans() {
        EditorBlanket.showEditor();
        setCollaborativeEditorStatus('off');
    }

    function makePromise(events, timeout) {
        var promise = $.Deferred();

        AJS.bind(events, function handler(e, data) {
            AJS.debug("confluence.editor.synchrony.event: " + e.type + "." + e.namespace);
            AJS.unbind(events, handler);
            promise.resolve(data);
        });

        if (timeout) {
            setTimeout(promise.reject, timeout, 'timeout');
        }

        return promise;
    }

    function makeInitPromise(events, dataProvider, conditionProvider, existingPromise) {
        var promise = existingPromise || $.Deferred();

        // If the editor is already initialised or we don't need any data (and config is already available) we should be good to go
        if (isEditorInitialised() || (!dataProvider && Util.isSynchronyEnabledAndSupported())) {
            return settlePromiseIfReady(promise, dataProvider);
        }

        AJS.bind(events, function handler(e, data) {
            AJS.debug('confluence.editor.event ' + e.type + "." + e.namespace);

            promise.always(function () {
                AJS.unbind(events, handler);
            });

            settlePromiseIfReady(promise, dataProvider, data, conditionProvider);
        });

        return promise;
    }

    function settlePromiseIfReady(promise, dataProvider, data, conditionProvider) {
        if (!Util.isSynchronyEnabledAndSupported()) {
            return promise.reject('synchrony-not-enabled');
        }

        if (!conditionProvider || conditionProvider()) {
            var result = dataProvider ? dataProvider(data) : data;
            return result && result.error ? promise.reject(result.error) : promise.resolve(result);
        }

        return promise;
    }

    Util.timeEnd("confluence.editor.synchrony.jsLoad");

    return exports;
});

require('confluence/module-exporter').safeRequire('confluence-collaborative-editor-plugin/synchrony-editor', function(SynchronyEditor) {
    SynchronyEditor.init();
});