define('confluence-collaborative-editor-plugin/synchrony-presence', [
    'backbone',
    'confluence-collaborative-editor-plugin/avatar-list-view',
    'confluence-collaborative-editor-plugin/overlay-view',
    'ajs',
    'confluence/meta',
    'confluence/templates',
    'jquery',
    'underscore'
], function(
    Backbone,
    AvatarListView,
    OverlayView,
    AJS,
    Meta,
    Templates,
    $,
    _
) {
    'use strict';

    var userSessions = new Backbone.Collection;
    var initialised = false;
    var currentUserOrigin = "";
    var tinymce;
    var contributors = [];

    // HACK! This only exists so tests can inject a mock tinymce. Apparently we can't mock inline requires with squire.
    // Note: tinymce needs to be required late to avoid issues with how it is loaded.
    function setTinyMce(newTinymce) {
        tinymce = newTinymce;
    }

    function appendPresenceTo(currentUser, entity, docId, rootContainer, lastKnownContributors) {
        // We always want to update this variable.
        currentUserOrigin = currentUser;

        var $presenceContainer = null;

        if (!initialised) {
            tinymce = tinymce || require('tinymce');

            if (!rootContainer) {
                AJS.log('synchronyPresence error: specify a valid container.');
                return;
            }

            initialised = true; // Prevent further initialisation.

            $presenceContainer = $(Templates.SynchronyPresence.container({
                docId: docId
            }));
            $(rootContainer).append($presenceContainer);

            new AvatarListView({
                el: "#avatar-list",
                collection: userSessions
            });

            var overlayView = new OverlayView({
                collection: userSessions
            });
            $(rootContainer).append(overlayView.render());

            // Add the last known contributors.
            contributors = lastKnownContributors;
            _renderContributors();

            AJS.bind('editor-heartbeat', function (e, data) {
                if (_.isArray(data.contributors)) {
                    contributors = data.contributors;
                    _renderContributors();
                }
            });

            entity.on('presence', function(event) {
                overlayView.hideInlineDialog();
                event.joined.filter(
                    function(userSession) {
                        return userSession.origin !== currentUserOrigin;
                    }
                ).forEach(
                    function(userSession) {
                        userSession.id = userSession.origin;
                        userSession.fullname = userSession.fullname || AJS.I18n.getText("anonymous.name");
                        userSession.name = userSession.name || "anonymous";
                        userSession.avatarURL = AJS.contextPath() + userSession.avatarURL;
                        userSession.active = true;
                        _removeAssociatedContributor(userSession);
                        userSessions.add(userSession, { at: _getActiveCount() });
                    }
                );

                event.left.forEach(function(userSession) {
                    userSession.id = userSession.origin;
                    userSessions.remove(userSession);
                    _renderContributors();
                });

                AJS.trigger('analyticsEvent', {
                    name: 'confluence.synchrony.user.in.session',
                    data: {
                        numOtherUsers: userSessions.length,
                        draftId: Meta.get('draft-id'),
                        contentId: Meta.get('content-id')
                    }
                });
            });

            // Hide inline dialog on editor interaction.
            tinymce.EditorManager.activeEditor.onClick.add(_.bind(OverlayView.prototype.hideInlineDialog, overlayView));
        }

        return $presenceContainer;
    }

    function _getActiveCount() {
        return userSessions.filter(function (c) { return c.get('active'); }).length;
    }

    function _removeAssociatedContributor(contributor) {
        var matching = userSessions.findWhere({ active: false, id: contributor.name });
        if (matching) {
            userSessions.remove(matching);
        }
    }

    function _renderContributors() {
        var toRemove = [];

        // Remove historical contributors from dom if they don't appear in the contributors array.
        userSessions.each(function (contributorDOM) {
            if (!(contributorDOM.get('active') || _.some(contributors, function (contributorReal) { return contributorDOM.get('id') === contributorReal.name }))) {
                toRemove.push(contributorDOM);
            }
        });

        // We can't remove these whilst iterating over the collection, so we remove them here.
        _.each(toRemove, function (user) {
            userSessions.remove(user);
        });

        // Insert historical contributors that aren't already in the DOM.
        _.each(contributors, function (contributorReal) {
            if (!(contributorReal.name === Meta.get('remote-user') || userSessions.any(
                    function (contributorDOM) {
                        return contributorDOM.get('id') === contributorReal.name || contributorDOM.get('name') === contributorReal.name
                    }
                ))
            ) {
                contributorReal.active = false;
                contributorReal.id = contributorReal.name;
                userSessions.add(contributorReal);
            }
        });
    }

    return {
        appendTo: appendPresenceTo,
        setTinyMce: setTinyMce
    };
});